/* Name: asp162bl.c
 * Project: AVR USB driver for AT90USB
 * Author: Koji Suzuki
 * Creation Date: 2009-12-16
 * Original Project: USBaspLoader
 * Original Author: Christian Starkjohann
 * Original Creation Date: 2007-12-08
 * Tabsize: 8
 * Copyright: (c) 2007 by OBJECTIVE DEVELOPMENT Software GmbH
 * Copyright: (c) 2009 by Koji Suzuki (suz@nmj2.sakura.ne.jp)
 * License: GNU GPL v2 (see License.txt)
 * Original Revision: Id: main.c 730 2009-03-20 09:05:11Z cs 
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
//#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#if 0
#include <usbdrv/usbdrv.h>
#include <usbconfig.h>
#else
#define USBDDR		DDRA
#define USBMINUS	PA7
#include <usbruntime.h>
#endif

static inline void busy_led_on(){}
static inline void busy_led_off(){}
static inline void connect_led_on(){}
static inline void connect_led_off(){}

#define TYPE_2313       0
#define TYPE_861        1
#define TYPE_T88A       2
#define TYPE_T88B       3
#define TYPE_T88C       4
#define TYPE_T88D       5
#define TYPE_T88E       6

static inline void bootLoaderCheckCondition() {
#ifdef KBD_TYPE
  #if (KBD_TYPE == TYPE_T88D) || (KBD_TYPE == TYPE_T88E)
	DDRC |= (1<<PC3);
	PORTC |= (1<<PC1); // pullup
	if (PINC & (1<<PC1) GPIOR0 |= 0xC0;
  #elif (KBD_TYPE == TYPE_T88C)
	DDRD |= (1<<PD7);
	PORTB |= (1<<PB1); // pullup
	if (PINB & (1<<PB1)) GPIOR0 |= 0xC0;
  #elif (KBD_TYPE == TYPE_T88B)
	DDRC |= (1<<PC3);
	PORTC |= (1<<PC1); // pullup
	if (PINC & (1<<PC1)) GPIOR0 |= 0xC0;
  #elif (KBD_TYPE == TYPE_T88A)
	DDRB |= (1<<PB7);
	PORTD |= (1<<PD6); // pullup
	if (PIND & (1<<PD6)) GPIOR0 |= 0xC0;
  #elif (KBD_TYPE == TYPE_861)
	DDRA = 0x77;
	PORTA = 0x65 | (1<<PA7); // pullup
	if (PINA & (1<<PA7)) GPIOR0 |= 0xC0;
  #endif
#else
  #if defined(__AVR_ATtiny44__)
	PORTA |= (1<<PA1); // pullup
	if (PINA & (1<<PA1)) GPIOR0 |= 0xC0;
  #endif
#endif
}
#define bootLoaderEnterCondition !(GPIOR0 & (1<<7))
#define bootLoaderExitCondition (GPIOR0 & (1<<7))

static inline void bootLoaderEnter() {
}
static inline void bootLoaderExit() {
#ifdef KBD_TYPE
  #if (KBD_TYPE == TYPE_T88D) || (KBD_TYPE == TYPE_T88E)
	DDRC &= ~(1<<PC3);
	PORTC &= ~(1<<PC1); // pullup
  #elif (KBD_TYPE == TYPE_T88C)
	DDRD &= ~(1<<PD7);
	PORTB &= ~(1<<PB1); // pullup
  #elif (KBD_TYPE == TYPE_T88B)
	DDRC &= ~(1<<PC3);
	PORTC &= ~(1<<PC1); // pullup
  #elif (KBD_TYPE == TYPE_T88A)
	DDRB &= ~(1<<PB7);
	PORTD &= ~(1<<PD6); // pullup
  #elif (KBD_TYPE == TYPE_861)
	DDRA = 0;
	PORTA = 0;
  #endif
#else
  #if defined(__AVR_ATtiny44__)
	PORTA &= ~(1<<PA1); // pullup
  #endif
#endif
}

#define USB_RQ_REQ	1
#define USB_RQ_VALUE_L	2
#define USB_RQ_VALUE_H	3
#define USB_RQ_INDEX_L	4
#define USB_RQ_INDEX_H	5
#define USB_RQ_LENGTH_L	6
#define USB_RQ_LENGTH_H	7

#ifndef BOOTLOADER_ADDRESS
#define BOOTLOADER_ADDRESS	(uint16_t)0x600
#endif
#ifndef BOOTLOADER_INT
#define BOOTLOADER_INT		(uint16_t)(1<<1)
#endif

#define HAVE_FUSE_ACCESS		1
#define HAVE_CHIP_ERASE			1
#define HAVE_SIGNATURE_ACCESS		0
#define HAVE_FLASH_BYTE_ACCESS		0
#define HAVE_EEPROM_BYTE_ACCESS		1
#define HAVE_EEPROM_PAGED_ACCESS	0

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define nop()	__asm__ volatile ("nop" )
#define	wdr()	__asm__ volatile("wdr" )

#define __cli()	__asm__ volatile ("cli" )
#define __sei()	__asm__ volatile ("sei" )

#ifndef APP_FLASHEND
#define APP_FLASHEND	(BOOTLOADER_ADDRESS -1)
#endif

/* ------------------------------------------------------------------------ */

/* Request constants used by USBasp */
#define USBASP_FUNC_CONNECT         1
#define USBASP_FUNC_DISCONNECT      2
#define USBASP_FUNC_TRANSMIT        3
#define USBASP_FUNC_READFLASH       4
#define USBASP_FUNC_ENABLEPROG      5
#define USBASP_FUNC_WRITEFLASH      6
#define USBASP_FUNC_READEEPROM      7
#define USBASP_FUNC_WRITEEEPROM     8
#define USBASP_FUNC_SETLONGADDRESS  9
#define USBASP_FUNC_SETISPSCK       10

#define ISPCMD_CHIP_ERASE	    0xac
#define ISPCMD_READ_SIGNATURE	    0x30
#define ISPCMD_READ_EEPROM	    0xa0
#define ISPCMD_WRITE_EEPROM	    0xc0

#define ISPCMD_READ_FLASH_L	    0x20
#define ISPCMD_READ_FLASH_H	    0x28

#define ISPCMD_READ_FUSE_L	    0x50	/* lower/ext */
#define ISPCMD_READ_FUSE_H	    0x58	/* upper/lockbit */

/* Block mode flags */
#define PROG_BLOCKFLAG_FIRST    1
#define PROG_BLOCKFLAG_LAST     2

/* Block mode data size */
#define USBASP_READBLOCKSIZE   200
#define USBASP_WRITEBLOCKSIZE  200

/* ------------------------------------------------------------------------ */

/* defaults if not in config file: */
#ifndef HAVE_EEPROM_PAGED_ACCESS
#   define HAVE_EEPROM_PAGED_ACCESS 0
#endif
#ifndef HAVE_EEPROM_BYTE_ACCESS
#   define HAVE_EEPROM_BYTE_ACCESS  0
#endif
#ifndef BOOTLOADER_CAN_EXIT
#   define  BOOTLOADER_CAN_EXIT     0
#endif

#ifndef HAVE_LOCK_WRITE
#   define HAVE_LOCK_WRITE 0
#endif
/* ------------------------------------------------------------------------ */
#if (FLASHEND > USHRT_MAX)
#   define CURRENT_ADDRESS  currentAddress.l
#   define addr_t           uint32_t
#else
#   define CURRENT_ADDRESS  currentAddress.w[0]
#   define addr_t           uint16_t
#endif

typedef union longConverter{
    addr_t  l;
    uint16_t  w[sizeof(addr_t)/2];
    uint8_t   b[sizeof(addr_t)];
}longConverter_t;

#if BOOTLOADER_CAN_EXIT
static uint8_t            requestBootLoaderExit;
#endif
static longConverter_t  currentAddress; /* in bytes */
static uint8_t            bytesRemaining;
static uint8_t            isLastPage;
#if HAVE_EEPROM_PAGED_ACCESS
static uint8_t            currentRequest;
#endif

#if !HAVE_SIGNATURE_ACCESS
static uint8_t  signatureBytes[4] PROGMEM = {
    SIGNATURE_0, SIGNATURE_1, SIGNATURE_2, 0
};
#endif

/* ------------------------------------------------------------------------ */


static inline void leaveBootloader()
{
    void (*nullVector)(void) __attribute__((__noreturn__)) = NULL;
    bootLoaderExit();
    __cli();

/* We must go through a global function pointer variable instead of writing
 *  ((void (*)(void))0)();
 * because the compiler optimizes a constant 0 to "rcall 0" which is not
 * handled correctly by the assembler.
 */
    nullVector();
}

/* ------------------------------------------------------------------------ */
#ifndef SPMEM // tiny88
#define SPMEM SELFPRGEN
#endif

#if HAVE_FUSE_ACCESS
/* addr == 0 : lockbit
   addr == 1 : fuse Low (FLB)
   addr == 2 : fuse Ext (FEB)
   addr == 3 : fuse High (FHB)
*/
static inline uint8_t read_fuse(uint16_t addr) {
	uint8_t r;
	__cli();
	SPMCSR |= (1<<SPMEN) | (1<<RFLB);
	r = pgm_read_byte_near(addr);
	__sei();
	return r;
}
#endif

#if HAVE_SIGNATURE_ACCESS
static inline uint8_t read_signature(uint16_t addr) {
	uint8_t r;
	__cli();
	SPMCSR |= (1<<SPMEN) | (1<<RSIG);
	r = pgm_read_byte_near(addr);
	__sei();
	return r;
}
#endif


#define __spm(reg, op, address, data)   	  \
 (__extension__({                                 \
     __asm__ __volatile__                         \
     (                                            \
         "movw  r0, %3\n\t"                       \
         "sts %0, %1\n\t"                         \
         "spm\n\t"                                \
         "clr  r1\n\t"                            \
         :                                        \
         : "i" (_SFR_MEM_ADDR(reg)),              \
           "r" ((uint8_t)(op)),                   \
           "z" ((uint16_t)(address)),             \
           "r" ((uint16_t)(data))                 \
         : "r0"                                   \
     );                                           \
 }))

static inline void __erase_flash_page(uint16_t addr) {
	__cli();
	__spm(SPMCSR, (1<<SPMEN)|(1<<PGERS),  addr, 0);
	__sei();
}

static inline void __fill_flash_page(uint16_t addr, uint16_t data) {
	__cli();
	__spm(SPMCSR, (1<<SPMEN),  addr, data);
	__sei();
}
static inline void __write_flash_page(uint16_t addr) {
	__cli();
	__spm(SPMCSR, (1<<SPMEN)|(1<<PGWRT),  addr, 0);
	__sei();
}



static inline void erase_flash_page(uint16_t addr, int do_write) {
	uint16_t r;
	SPMCSR |= (1<<CTPB);
	if (addr < SPM_PAGESIZE) {
		r = pgm_read_word(0);
		__fill_flash_page(0, r);
		r = pgm_read_word(BOOTLOADER_INT);
		__fill_flash_page(BOOTLOADER_INT, r);
	}
	__erase_flash_page(addr);
	if (addr < SPM_PAGESIZE) {
#if 0 // needless ??
		for (r=0; r< SPM_PAGESIZE; r+=2)
			__fill_flash_page(r, 0xffff);
#endif
	    if (do_write)
		__write_flash_page(addr);
	}
}

uint16_t pend_vect0;
uint16_t pend_vect1;

static inline void fill_flash_page(uint16_t addr, uint16_t data) {
	if (addr == 0) {
		pend_vect0 = data;
	} else if (addr == BOOTLOADER_INT) {
		pend_vect1 = data;
	} else
		__fill_flash_page(addr, data);
}

static inline void clear_pending() {
    pend_vect0 = 0;
    pend_vect1 = 0;
}

static inline void fix_pending() {
    uint16_t a, r;
    if (pend_vect0 || pend_vect1) {
      SPMCSR |= (1<<CTPB);
      for (a= BOOTLOADER_ADDRESS; a< BOOTLOADER_ADDRESS+SPM_PAGESIZE; a+=2) {
	if (pend_vect0 && (a == BOOTLOADER_ADDRESS+4)) {
	    r = pend_vect0 - ((BOOTLOADER_ADDRESS+4 -0)>>1) ;
	    r = (r & 0xfff) | (pend_vect0 & 0xf000);
	} else if (pend_vect1 && (a == BOOTLOADER_ADDRESS+10)) {
	    r = pend_vect1 - ((BOOTLOADER_ADDRESS+10 -BOOTLOADER_INT)>>1) ;
	    r = (r & 0xfff) | (pend_vect1 & 0xf000);
	} else 
	    r = pgm_read_word_near(a);
	__fill_flash_page(a, r);
      }
      __erase_flash_page(BOOTLOADER_ADDRESS);
      __write_flash_page(BOOTLOADER_ADDRESS);
    }
    clear_pending();
}


static uint16_t read_flash(uint16_t addr) {
	uint16_t r,data;
	if (addr == 0) {
		data = pgm_read_word_near(BOOTLOADER_ADDRESS+4);
		r = data + ((BOOTLOADER_ADDRESS+4 -0)>>1);
	} else if (addr == BOOTLOADER_INT) {
		data = pgm_read_word_near(BOOTLOADER_ADDRESS+10);
		r = data + ((BOOTLOADER_ADDRESS+10 -BOOTLOADER_INT)>>1) ;
	} else if (addr >= BOOTLOADER_ADDRESS) {
		return 0xffff;
	} else {
		return pgm_read_word_near(addr & 0xfffe);
	}
	return (r & 0x0fff) | (data & 0xf000);
}


static inline uint8_t eeprom_read_byte(void *address) {
	while (EECR & (1<<EEPE))
		;
	EEAR = address;
	EECR |= (1<<EERE);
	return EEDR;
}

static inline void  eeprom_write_byte(void *address, uint8_t data) {
	while (EECR & (1<<EEPE))
		;
	EEAR = address;
	EEDR = data;
	__cli();
	EECR |= (1<<EEMPE);
	EECR |= (1<<EEPE);
	__sei();
}

uint8_t asp_setup(uint8_t *in, uint8_t *out)
{
    uint8_t  len = 0;
    uint8_t  request = in[USB_RQ_REQ]; 
    uint8_t  value_l = in[USB_RQ_VALUE_L];
#define  ispcmd	value_l
    uint8_t  value_h = in[USB_RQ_VALUE_H];
    uint8_t  rval = 0;
    uint16_t address;
    uint16_t wdata;

    if(request == USBASP_FUNC_TRANSMIT){   /* emulate parts of ISP protocol */
        address = (uint16_t)(value_h <<8) | in[USB_RQ_INDEX_L];
        if(ispcmd == ISPCMD_READ_SIGNATURE){
            rval = in[USB_RQ_INDEX_L] & 3;
#if HAVE_SIGNATURE_ACCESS
	    if (rval == 3) {
		    rval = 1;
	    } else {
		    rval <<= 1;
	    }
            rval = read_signature(rval);
#else
            rval = pgm_read_byte_near(signatureBytes+rval);
#endif
#if HAVE_EEPROM_BYTE_ACCESS
        }else if(ispcmd == ISPCMD_READ_EEPROM){
            rval = eeprom_read_byte((void *)address);
        }else if(ispcmd == ISPCMD_WRITE_EEPROM){
            eeprom_write_byte((void *)address, in[USB_RQ_INDEX_H]);
#endif
#if HAVE_FLASH_BYTE_ACCESS
        }else if((ispcmd & 0xf7) == ISPCMD_READ_FLASH_L) {
	    wdata = read_flash(address <<1);
	    if (ispcmd & 0x08) {  /* read FLASH upper */
		rval = wdata >> 8;
	    } else {
            	rval = wdata;
	    }
#endif
#if HAVE_FUSE_ACCESS
        }else if((ispcmd & 0xf7) == ISPCMD_READ_FUSE_L) {
	    rval = 0;
	    if (ispcmd & 0x08) rval |= 1;
	    if (value_h & 0x08) rval |= 2;
	    rval = read_fuse(rval);
#endif
#if HAVE_CHIP_ERASE
        }else if(ispcmd == ISPCMD_CHIP_ERASE) {
          if(value_h == 0x80){
            addr_t addr;
            for(addr = 0; addr < (APP_FLASHEND) + 1; addr += SPM_PAGESIZE) {
                /* erase page and wait */
                erase_flash_page(addr, 1);
            }
          }
#endif
        }else{
            /* ignore all others, return default value == 0 */
        }
        out[3] = rval;
        len = 4;
    }else if(request == USBASP_FUNC_ENABLEPROG || request == USBASP_FUNC_SETISPSCK){
        out[0] = 0;
        len = 1;
    }else if(request == USBASP_FUNC_CONNECT){
#if BOOTLOADER_CAN_EXIT
#else
	connect_led_on();
#endif
    }else if(request == USBASP_FUNC_DISCONNECT){
#if BOOTLOADER_CAN_EXIT
        requestBootLoaderExit = 1;      /* allow proper shutdown/close of connection */
#else
	connect_led_off();
#endif
    }else if(request == USBASP_FUNC_SETLONGADDRESS){
#if (FLASHEND > USHRT_MAX)
            currentAddress.b[2] = in[USB_RQ_INDEX_L];
            currentAddress.b[3] = in[USB_RQ_INDEX_H];
#endif
    }else if(request > USBASP_FUNC_SETISPSCK) {
    }else {
        currentAddress.b[0] = value_l;
        currentAddress.b[1] = value_h;
	len = in[USB_RQ_LENGTH_L];
	if(request == USBASP_FUNC_READEEPROM || request == USBASP_FUNC_READFLASH) {
	} else {
	    clear_pending();
            /* if(in[USB_RQ_REQ] == USBASP_FUNC_WRITEFLASH) only evaluated during writeFlash anyway */
            isLastPage = in[USB_RQ_INDEX_H] & PROG_BLOCKFLAG_LAST;
#if HAVE_EEPROM_PAGED_ACCESS
            currentRequest = request;
#endif
	    bytesRemaining = len;
            len = 0; // NO ERROR @ OUT
	}
    }
    busy_led_off();
    return len;
#undef ispcmd
}

uint8_t asp_in(uint8_t *data, uint8_t len)
{
    uint8_t i;
#if HAVE_EEPROM_PAGED_ACCESS
    if (currentRequest == USBASP_FUNC_READEEPROM) {
        for (i = 0; i < len; i++) {
            *data++ = eeprom_read_byte((void *)CURRENT_ADDRESS++);
        }
    } else
#endif
    { // USBASP_FUNC_READFLASH
        for (i = 0; i < len; i+=2) {
            *(uint16_t *)data = read_flash(CURRENT_ADDRESS);
            CURRENT_ADDRESS += 2;
	    data += 2;
        }
    }
}

uint8_t asp_out(uint8_t *data, uint8_t len)
{
    uint8_t isLast;
    uint8_t i;

    busy_led_on();
    if(len > bytesRemaining)
        len = bytesRemaining;
    bytesRemaining -= len;
    isLast = (bytesRemaining == 0);
#if HAVE_EEPROM_PAGED_ACCESS
    if(currentRequest == USBASP_FUNC_WRITEEEPROM){
        for(i = 0; i < len; i++){
            eeprom_write_byte((void *)CURRENT_ADDRESS++, *data++);
        }
    }else{ // USBASP_FUNC_WRITEFLASH
#endif
        for(i = 0; i < len;){
            if((CURRENT_ADDRESS & (SPM_PAGESIZE - 1)) == 0){    /* if page start: erase */
                erase_flash_page(CURRENT_ADDRESS, 0);   /* erase page */
            }
            i += 2;
            fill_flash_page(CURRENT_ADDRESS, *(uint16_t *)data);
            CURRENT_ADDRESS += 2;
            data += 2;
            /* write page when we cross page boundary or we have the last partial page */
            if((CURRENT_ADDRESS & (SPM_PAGESIZE - 1)) == 0 
			|| (isLast && (i >= len) && isLastPage)){

                __write_flash_page(CURRENT_ADDRESS - 2);
		fix_pending();
            }
        } 
#if HAVE_EEPROM_PAGED_ACCESS
    }
#endif
    busy_led_off();
    return isLast?(uint8_t)255:0; 
}
/* ------------------------------------------------------------------------ */

extern usbMsgLen_t  usbFunctionDescriptor(struct usbRequest *rq);
void main(void)
{
uchar   i;
#ifdef F_SCALER
    uint8_t save_clkpr;

    save_clkpr = CLKPR;
    /* set system_clock pre scaler 1/8 => 1/1 */
    CLKPR = (1<<CLKPCE);
    CLKPR = F_SCALER;
#endif
    bootLoaderCheckCondition();
    if(!bootLoaderEnterCondition) goto leave;
    wdt_enable(WDTO_1S);
    _usbFunctionDescriptor = usbFunctionDescriptor;
    _usbFunctionSetup = asp_setup;
    _usbFunctionRead = asp_in;
    _usbFunctionWrite = asp_out;

    usbInit();
    usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
    i = 0;
    while(--i){             /* fake USB disconnect for > 250 ms */
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();
    __sei();
    if(bootLoaderEnterCondition){
    	bootLoaderEnter();
        do{
	    wdt_reset();
            usbPoll();
#if BOOTLOADER_CAN_EXIT
            if(requestBootLoaderExit){
                if(--i == 0){
                    if(--j == 0)
                        break;
                }
            }
#endif
        }while(!bootLoaderExitCondition);  /* main event loop */
    }
    usbDeviceDisconnect();
#ifdef F_SCALER
    CLKPR = (1<<CLKPCE);
    CLKPR = save_clkpr;
#endif
leave:
    leaveBootloader();
//    return 0;
}

/* ------------------------------------------------------------------------- */
