/* Name: usbdrv.c
 * Project: V-USB, virtual USB port for Atmel's(r) AVR(r) microcontrollers
 * Author: Christian Starkjohann
 * Creation Date: 2004-12-29
 * Tabsize: 4
 * Copyright: (c) 2005 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt), GNU GPL v3 or proprietary (CommercialLicense.txt)
 * This Revision: $Id: usbdrv.c 763 2009-08-22 10:27:24Z cs $
 */

//#include "usbportability.h"
#include <stdio.h>
//#include <usbdrv/usbdrv.h>
#include <usbruntime.h>

/*
General Description:
This module implements the C-part of the USB driver. See usbdrv.h for a
documentation of the entire driver.
*/
/*
	0080011f A __stack
	00800120 B usbMsgPtr
	00800122 B _usbFunctionSetup
	00800124 B _usbFunctionDescriptor
	00800126 B _usbFunctionWrite
	00800128 B _usbFunctionRead
	0080012a B usbRxToken
	0080012b B usbConfiguration
	0080012c B usbRxBuf
	00800142 B usbInputBufOffset
	00800143 B usbDeviceAddr
	00800144 B usbNewDeviceAddr
	00800145 B usbRxLen
	00800146 B usbCurrentTok
	00800147 B usbTxLen
	00800148 B usbTxBuf
	00800153 B usbMsgLen
	00800154 B usbMsgFlags
*/
/* ------------------------------------------------------------------------- */
#define API_SECTION(addr)  __attribute__ ((section(".api"))) 
//#define API_SECTION(addr)  __attribute__ ((address(API_ADDRESS+addr)))
//#define API_SECTION(addr)  __attribute__ ((section(".noinit"))) 
//#define API_SECTION(addr)  __attribute__ ((section(".noinit"),address(API_ADDRESS+addr)))

usbMsgLen_t (* _usbFunctionDescriptor)(struct usbRequest *rq) API_SECTION(0x06);
usbMsgLen_t (* _usbFunctionSetup)(uchar data[8]) API_SECTION(0x04);
uchar (* _usbFunctionWrite)(uchar *data, uchar len) API_SECTION(0x08);
uchar (* _usbFunctionRead)(uchar *data, uchar len) API_SECTION(0x0a);


/* raw USB registers / interface to assembler code: */
uchar usbRxBuf[2*USB_BUFSIZE] API_SECTION(0x0e);  /* raw RX buffer: PID, 8 bytes data, 2 bytes CRC */
uchar       usbInputBufOffset API_SECTION(0x24);  /* offset in usbRxBuf used for low level receiving */
uchar       usbDeviceAddr API_SECTION(0x25);      /* assigned during enumeration, defaults to 0 */
uchar       usbNewDeviceAddr API_SECTION(0x26);   /* device ID which should be set after status phase */
uchar       usbConfiguration API_SECTION(0x0d);   /* currently selected configuration. Administered by driver, but not used */
volatile schar usbRxLen API_SECTION(0x27);        /* = 0; number of bytes in usbRxBuf; 0 means free, -1 for flow control */
uchar       usbCurrentTok API_SECTION(0x28);      /* last token received or endpoint number for last OUT token if != 0 */
uchar       usbRxToken API_SECTION(0x0c);         /* token for data we received; or endpont number for last OUT */
volatile uchar usbTxLen API_SECTION(0x00); /*  = USBPID_NAK;   /* number of bytes to transmit with next IN token or handshake token */
uchar       usbTxBuf[USB_BUFSIZE] API_SECTION(0x29);/* data to transmit with next IN, free if usbTxLen contains handshake token */
#if USB_COUNT_SOF
volatile uchar  usbSofCount API_SECTION(0xff);    /* incremented by assembler module every SOF */
#endif
#if USB_CFG_HAVE_INTRIN_ENDPOINT && !USB_CFG_SUPPRESS_INTR_CODE
usbTxStatus_t  usbTxStatus1 API_SECTION(0xff);
#   if USB_CFG_HAVE_INTRIN_ENDPOINT3
usbTxStatus_t  usbTxStatus3 API_SECTION(0xff);
#   endif
#endif
#if USB_CFG_CHECK_DATA_TOGGLING
uchar       usbCurrentDataToken API_SECTION(0xff);/* when we check data toggling to ignore duplicate packets */
#endif

/* USB status registers / not shared with asm code */
uchar *usbMsgPtr API_SECTION(0x02);     /* data to transmit next -- ROM or RAM address */
usbMsgLen_t  usbMsgLen API_SECTION(0x01); /* = USB_NO_MSG; /* remaining number of bytes */
uchar        usbMsgFlags API_SECTION(0x34);    /* flag values see below */

